/*
 *  CentipedeStory.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.geom.*;
import java.util.*;

// controlling class for the game
public class CentipedeStory extends Story {

  // event for bombs exploding
  static public class EventExplosion extends StoryEvent {
    public int mNumDead;
    public EventExplosion(int n) { mNumDead = n; }
    public int numDead() { return mNumDead; }
  } // class EventExplosion
  
  // event for bombs landing safely
  static public class EventLanding extends StoryEvent {
    public EventLanding() {}
  } // class EventLanding
  
  // positions of text
  private static final int kTopYPos      = 15,
                           kIncomingXPos = 10,
                           kLivesXPos    = Env.screenWidth()-10;
  private static final int kTextHeight   = 20,
                           kTextFontSize = 14,
                           kTextLayer    = 10;
  
  // game over message
  private static final int kGameOverHeight = 50,
                           kGameOverYPos   = Env.screenHeight()/2 - 20,
                           kGameOverLayer  = 100;
  private static final int kGameOverDelay  = 30;
  private static final int kRestartDelay   = 100;
  
  // title text
  private static final int kTitle1Height = 50,
                           kTitle1YPos   = Env.screenHeight()/2 - 30,
                           kTitle2Height = 20,
                           kTitle2YPos   = Env.screenHeight()/2 - 0,
                           kTitle3Height = 24,
                           kTitle3YPos   = Env.screenHeight() - 22,
                           kTitleLayer   = 100;
  private static final int kStartDelay   = 1;
  
  // instruction text
  private static final int kHelpHeight = 20,
                           kHelpYPos   = Env.screenHeight() - 20,
                           kHelpLayer  = 100;
  
  // comment text
  private static final int kCommentHeight = 40,
                           kCommentYPos   = Env.screenHeight()/2 - 20,
                           kCommentLayer  = 100,
                           kCommentTime1  = 260,
                           kCommentTime2  = 130,
                           kCommentTime3  = 60,
                           kCommentTime4  = -10;
  
  // parameters for different levels
  private static final int kLevelNumIncoming[] = { 8, 15, 20, 25, 35, 40 },
                           kLevelNumLives[]    = { 5, 10, 10, 10, 10, 10 }; 
  
  // when to blow up the flower
  private static final int kStartSelfDestruct  = 10,
                           kNextLevelDelay     = 60,
                           kRemoveCommentDelay = 60;
  
  // points
  private static final int kScoreForLanding    = 20,
                           kScoreForCompleting = 50;
  
  // references to various sprites
  private Chain      mPlayer;
  private Background mBackground;
  private Fence      mFence;
  
  // reference to various bits of text
  private TextImage mLivesText = null;
  private TextImage mIncomingText = null;
  private TextImage mScoreText = null;
  private TextImage mGameOverText = null;
  private TextImage mHelpText = null;
  private TextImage mCommentText = null;
  private TextImage mTitleText1 = null,
                    mTitleText2 = null,
                    mTitleText3 = null;
  
  // counters
  private int mNumLives,
              mNumIncoming,
              mScore,
              mHighScore;
  
  // true when the bombs have reached their target
  private boolean mSelfDestructTriggered;
  
  // current level (1, 2, 3, ...)
  private int mLevel;
  
  // timers
  private int mIncomingTimer,
              mGameOverTimer,
              mRestartTimer,
              mNewLevelTimer,
              mRemoveCommentTimer;
  
  // prepare resources
  static public void initialize() {

    Explosion.initialize();
    Flower.initialize();
    Incoming.initialize();
    Segment.initialize();
    
  } // initialize()

  // constructor
  public CentipedeStory() {

    initialize();

  } // constructor

  // advance the Story by one frame
  public Story advance(LinkedList<StoryEvent> storyEvents,
                       SpriteManager spriteManager) {

    // process the story event list
    for ( Iterator<StoryEvent> it = storyEvents.iterator() ; it.hasNext() ; ) {
      StoryEvent event = it.next();

      if ( event instanceof Story.EventGameBegins ) {
        // first frame of the story, so set everything up
        prepareNewStory(spriteManager);
        it.remove();
      } // EventGameBegins

      if ( event instanceof EventExplosion ) {
        if ( !mSelfDestructTriggered ) {
          setLives(mNumLives - ((EventExplosion)event).numDead(),
                   spriteManager);
        }
        it.remove();
      } // EventExplosion
      
      if ( event instanceof EventLanding ) {
        if ( mNumLives > 0  ) {
          setScore(mScore+kScoreForLanding, spriteManager);
        }
        it.remove();
      } // EventLanding
      
    } // for each story event

    // add bombs
    if ( mPlayer.controllable() && 
         mNumIncoming > 0 && --mIncomingTimer <= 0 ) {
      throwBomb(spriteManager);
      mIncomingTimer = 60;
      setIncoming(mNumIncoming-1, spriteManager);
    }
    
    // no lives left => game over
    if ( mNumLives == 0 && mPlayer.controllable() ) {
      mPlayer.disableControl();
      mGameOverTimer = kGameOverDelay;
    }
    
    // none incoming => level complete
    if ( mNumIncoming == 0 && mNumLives > 0 && mPlayer.controllable() &&
         spriteManager.findSpriteOfType(Incoming.class) == null &&
         spriteManager.findSpriteOfType(Explosion.class) == null ) {
      if ( mHelpText != null ) {
        spriteManager.removeSprite(mHelpText);
        mHelpText = null;
      }
      mPlayer.disableControl();
      spriteManager.addSprite(new Flower(mLevel));
    }
    
    // wait to destroy flower, then wait to start next level
    if ( mNumIncoming == 0 && mNumLives > 0 && !mPlayer.controllable() ) {
      Flower f = (Flower)spriteManager.findSpriteOfType(Flower.class);
      if ( f != null ) {
        int time = f.timeTillDetach(); 
        if ( time == kCommentTime1 ) {
          mCommentText = new TextImage("Target sighted!",
                                       Env.screenWidth()/2, kCommentYPos,
                                       kCommentHeight, Color.WHITE, 
                                       kTextFontSize, 0, kCommentLayer);
          spriteManager.addSprite(mCommentText);
        } else if ( time == kCommentTime2 ) {
          spriteManager.removeSprite(mCommentText);
        } else if ( time == kCommentTime3 ) {
          mCommentText = new TextImage("Engage the petally foe!",
                                       Env.screenWidth()/2, kCommentYPos,
                                       kCommentHeight, Color.WHITE, 
                                       kTextFontSize, 0, kCommentLayer);
          spriteManager.addSprite(mCommentText);
        } else if ( time == kStartSelfDestruct ) {
          if ( mLivesText != null ) {
            spriteManager.removeSprite(mLivesText);
            mLivesText = null;
          }
          if ( mIncomingText != null ) {
            spriteManager.removeSprite(mIncomingText);
            mIncomingText = null;
          }
          mSelfDestructTriggered = true;
          mPlayer.selfDestruct();
          setScore(mScore+kScoreForCompleting*mPlayer.size(), spriteManager);
          mNewLevelTimer = kNextLevelDelay;
        } else if ( time == kCommentTime4 ) {
          spriteManager.removeSprite(mCommentText);
          mCommentText = null;          
        }
      }
    }
    
    // start a new level
    if ( mNewLevelTimer > 0 && mPlayer.size() == 0 &&
         spriteManager.findSpriteOfType(Explosion.class) == null ) {
      if ( --mNewLevelTimer == 0 ) {
        mLevel += 1;
        mCommentText = new TextImage("Level " + mLevel,
                                     Env.screenWidth()/2, kCommentYPos,
                                     kGameOverHeight, Color.WHITE, 
                                     kTextFontSize, 0, kTextLayer);
        mRemoveCommentTimer = kRemoveCommentDelay;
        spriteManager.addSprite(mCommentText);
        prepareNewLevel(spriteManager);
      }
    }
    
    // show game over message
    if ( mGameOverTimer > 0 ) {
      if ( --mGameOverTimer == 0 ) {
        if ( mHelpText != null ) {
          spriteManager.removeSprite(mHelpText);
          mHelpText = null;
        }
        mGameOverText = new TextImage("Game Over",
                                      Env.screenWidth()/2, kGameOverYPos,
                                      kGameOverHeight, Color.WHITE, 
                                      kTextFontSize, 0, kGameOverLayer);
        spriteManager.addSprite(mGameOverText);
        mRestartTimer = kRestartDelay;
      }
    }
    
    // show titles
    if ( mRestartTimer > 0  ) {
      if ( spriteManager.findSpriteOfType(Incoming.class) == null &&
           spriteManager.findSpriteOfType(Explosion.class) == null &&
           --mRestartTimer == 0 ) {
        if ( mGameOverText != null ) {
          spriteManager.removeSprite(mGameOverText);
          mGameOverText = null;
        }
        if ( mLivesText != null ) {
          spriteManager.removeSprite(mLivesText);
          mLivesText = null;
        }
        if ( mIncomingText != null ) {
          spriteManager.removeSprite(mIncomingText);
          mIncomingText = null;
        }
        if ( mScoreText != null || mHighScore > 0 ) {
          setHighScore(mScore, spriteManager);
        }
        mTitleText1 = new TextImage("End of Daisies",
                                    Env.screenWidth()/2, kTitle1YPos,
                                    kTitle1Height, Color.WHITE, 
                                    kTextFontSize, 0, kTitleLayer);
        mTitleText2 = new TextImage("dishmoth@yahoo.co.uk",
                                    Env.screenWidth()/2, kTitle2YPos,
                                    kTitle2Height, Color.WHITE, 
                                    kTextFontSize, 0, kTitleLayer);
        mTitleText3 = new TextImage("Click to Play",
                                    Env.screenWidth()/2, kTitle3YPos,
                                    kTitle3Height, Color.WHITE, 
                                    kTextFontSize, 0, kTitleLayer);
        spriteManager.addSprite(mTitleText1);
        spriteManager.addSprite(mTitleText2);
        spriteManager.addSprite(mTitleText3);
        Env.mouse().reset();
      }
    }
    
    // remove the title screen and start the game
    if ( mTitleText1 != null && Env.mouse().clicked() ) {
      if ( mTitleText1 != null ) spriteManager.removeSprite(mTitleText1);
      if ( mTitleText2 != null ) spriteManager.removeSprite(mTitleText2);
      if ( mTitleText3 != null ) spriteManager.removeSprite(mTitleText3);
      mTitleText1 = mTitleText2 = mTitleText3 = null;
      mLevel = 1;
      prepareNewLevel(spriteManager);
    }
    
    // dispose of any comment objects
    if ( mRemoveCommentTimer > 0 ) {
      if ( --mRemoveCommentTimer == 0 ) {
        if ( mCommentText != null ) {
          spriteManager.removeSprite(mCommentText);
          mCommentText = null;
        }
      }
    }
    
    // no change of story
    return null;

  } // advanceGame()

  // things to do on the first frame
  private void prepareNewStory(SpriteManager spriteManager) {

    mBackground = new Background();
    spriteManager.addSprite(mBackground);
    
    mFence = new Fence(40);
    spriteManager.addSprite(mFence);
    
    mPlayer = new Chain();
    mPlayer.disableControl();
    spriteManager.addSprite(mPlayer);

    mRestartTimer = kStartDelay;
    mLevel = 1;
    mScore = mHighScore = 0;
    
    mSelfDestructTriggered = false;

  } // prepareNewStory()

  // things to do for a new game level
  private void prepareNewLevel(SpriteManager spriteManager) {
    
    assert( mPlayer != null );
    mPlayer.resetSegments();
    mPlayer.enableControl();
    
    mIncomingTimer = ((mLevel == 1) ? 60 : 90);
    mSelfDestructTriggered = false;
    
    assert( mLevel > 0 );
    int index1 = Math.min(mLevel-1, kLevelNumLives.length-1);  
    setLives(kLevelNumLives[index1], spriteManager);
    
    int index2 = Math.min(mLevel-1, kLevelNumIncoming.length-1);  
    setIncoming(kLevelNumIncoming[index2], spriteManager);
    
    if ( mLevel == 1 ) setScore(0, spriteManager);
    
    if ( mLevel == 1 ) {
      mHelpText = new TextImage("Click with the left and right mouse buttons "
                                + "to avoid the falling bombs",
                                Env.screenWidth()/2, kHelpYPos,
                                kHelpHeight, Color.WHITE, 
                                kTextFontSize, 0, kHelpLayer);
      spriteManager.addSprite(mHelpText);
    }
    
  } // prepareNewGame
  
  // introduce a new bomb
  private void throwBomb(SpriteManager spriteManager) {

    int minWidth = (int)Math.round(0.3*Env.screenWidth());
    int farLeft  = (int)Math.round(0.1*Env.screenWidth()),
        farRight = (int)Math.round(0.9*Env.screenWidth());
    
    int xMin = mPlayer.minXPos(),
        xMax = mPlayer.maxXPos();
    int width = xMax - xMin;
    
    if ( width < minWidth ) {
      int d = minWidth - width;
      xMin -= d/2;
      xMax += d/2;
      width = minWidth;
    }    
    if ( xMax > farRight ) {
      int d = xMax - farRight;
      xMax -= d;
      xMin -= d;
    }
    if ( xMin < farLeft ) {
      int d = farLeft - xMin;
      xMax += d;
      xMin += d;
      xMax = Math.min(xMax, farRight);
    }
    
    int xTarget = Env.randomInt(xMin, xMax);
    spriteManager.addSprite(new Incoming(xTarget));

  } // throwBomb()
  
  // change the number of incoming bombs
  private void setIncoming(int num, SpriteManager spriteManager) {
    
    if ( mNumIncoming == num ) return;
    
    mNumIncoming = Math.max(0, num);

    if ( mIncomingText != null ) spriteManager.removeSprite(mIncomingText);
    mIncomingText = new TextImage("Incoming " + mNumIncoming, 
                                  kIncomingXPos, kTopYPos, kTextHeight,
                                  Color.WHITE, kTextFontSize, 
                                  -1, kTextLayer);
    spriteManager.addSprite(mIncomingText);
    
  } // setIncoming()
  
  // change the number of lives
  private void setLives(int num, SpriteManager spriteManager) {
    
    if ( mNumLives == num ) return;
    
    mNumLives = Math.max(0, num);

    if ( mLivesText != null ) spriteManager.removeSprite(mLivesText);
    mLivesText = new TextImage("Lives " + mNumLives, 
                               kLivesXPos, kTopYPos, kTextHeight,
                               Color.WHITE, kTextFontSize, 
                               +1, kTextLayer);
    spriteManager.addSprite(mLivesText);
    
  } // setLives()
  
  // change the score
  private void setScore(int num, SpriteManager spriteManager) {
    
    mScore = Math.max(0, num);

    if ( mScoreText != null ) spriteManager.removeSprite(mScoreText);
    mScoreText = new TextImage("Score " + mScore, 
                               Env.screenWidth()/2, kTopYPos, kTextHeight,
                               Color.WHITE, kTextFontSize, 
                               0, kTextLayer);
    spriteManager.addSprite(mScoreText);
    
  } // setScore()
  
  // change the high score
  private void setHighScore(int num, SpriteManager spriteManager) {
    
    mHighScore = Math.max(mHighScore, num);

    if ( mScoreText != null ) spriteManager.removeSprite(mScoreText);
    mScoreText = new TextImage("High Score " + mHighScore, 
                               Env.screenWidth()/2, kTopYPos, kTextHeight,
                               Color.WHITE, kTextFontSize, 
                               0, kTextLayer);
    spriteManager.addSprite(mScoreText);
    
  } // setHighScore()
  
  // remove all sprites not needed for the next story
  private void tidySpriteList(SpriteManager spriteManager) {

    LinkedList<Sprite> deadSprites = new LinkedList<Sprite>();
    for (Sprite s : spriteManager.list()) {
      deadSprites.add(s);
    }
    if (deadSprites.size() > 0) spriteManager.removeSprites(deadSprites);

  } // tidySpriteList()

} // class CentipedeStory
