/*
 *  Chain.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;

// a collection of body segment
public class Chain extends Sprite {

  // how sprite is drawn relative to other sprites
  static private final int kScreenLayer = 50;

  //
  static private final int kInitialNumSegments = 3;
  
  // distance between segments in a chain
  static private final int kSegmentSpacing = 16;
  
  // animation of the segments
  static private final int kSegmentDelay = 1;

  // how near the mouse has to be
  static private final int kMouseMargin = Segment.width()/2 + 8;

  // detonate the segments
  static private final int kSelfDestructDelay = 5;
  
  // the segments
  private Vector<Segment> mSegments;
  
  // flags indicating contact between pairs of segments
  private Vector<Boolean> mChains;
  
  // mouse ready
  private boolean mMouseReady;
  
  // whether controls are active
  private boolean mPlayable;
  
  // index of segment to highlight (or -1)
  private int mHighlightIndex;
  
  // time for animation
  private int mTime;
  
  // time till next deliberate explosion (or 0)
  private int mSelfDestructTime;
  
  // constructor
  public Chain() {
    
    super(kScreenLayer);
    Segment.initialize();
    
    resetSegments();
    
    mMouseReady = false;
    mPlayable = true;
    
    mTime = 0;
    mSelfDestructTime = 0;
    
  } // constructor
  
  // put segments to initial state
  public void resetSegments() {

    if ( mSegments == null ) mSegments = new Vector<Segment>();
    if ( mChains == null ) mChains = new Vector<Boolean>();

    while ( mSegments.size() > 0 ) removeSegment(0);
    
    for ( int k = 0 ; k < kInitialNumSegments ; k++ ) {
      int x = Env.screenWidth()/2 - Segment.width()/2
            + (kSegmentSpacing*(2*k - (kInitialNumSegments-1)))/2;
      mSegments.add(new Segment(x));
    }
    for ( int k = 0 ; k < kInitialNumSegments-1 ; k++ ) {
      mChains.add(true);
    }
    
    mSelfDestructTime = 0;
    
  } // resetSegments()
  
  // enable or disable controls
  public void enableControl() { mPlayable = true; }
  public void disableControl() { mPlayable = false; }
  public boolean controllable() { return mPlayable; }
  
  // blow up all the segments
  public void selfDestruct() {
    
    mSelfDestructTime = 1; 
    mPlayable = false;
    
  } // selfDestruct()
  
  // range of segment positions
  public int minXPos() {
    
    if ( mSegments.size() == 0 ) return Env.screenWidth()/2;
    Segment seg = mSegments.get(0);
    return (int)Math.floor( seg.xPos() );
    
  } // minXPos()
  public int maxXPos() {
    
    if ( mSegments.size() == 0 ) return Env.screenWidth()/2;
    Segment seg = mSegments.get( mSegments.size()-1 );
    return (int)Math.ceil( seg.xPos() + Segment.width() );
    
  } // maxXPos()
  
  // number of segments
  public int size() { return mSegments.size(); }
  
  // insert a new segment
  public void addSegment(int xPos, double xVel) {
    
    int index;
    for ( index = 0 ; index < mSegments.size() ; index++ ) {
      Segment seg = mSegments.get(index);
      if ( seg.xPos() > xPos ) break;
    }
    mSegments.add(index, new Segment(xPos, xVel));
    if ( mSegments.size() > 1 ) {
      mChains.add(Math.min(index,mChains.size()), false);
    }
    
  } // addSegment()
  
  // discard a segment
  private void removeSegment(int index) {
    
    mSegments.remove(index);
    
    if ( index < mChains.size() ) {
      mChains.remove(index);
      if ( index > 0 ) mChains.set(index-1, false);
    } else if ( mChains.size() > 0 ){
      mChains.remove(index-1);
    }
    
  } // removeSegment()
  
  // check for collision of segments against a circle (returns true if so)
  public int collides(double x, double y, double r, boolean damage) {
    
    Vector<Integer> deathList = new Vector<Integer>();
    
    int numHit = 0;
    for ( int k = 0 ; k < mSegments.size() ; k++ ) {
      Segment seg = mSegments.get(k);
      boolean hit = seg.collides(x, y, r); 
      if ( hit ) { 
        numHit += 1;
        if ( damage ) deathList.add(k);
      }
    }
    
    for ( int k = deathList.size()-1 ; k >= 0 ; k-- ) {
      removeSegment(deathList.get(k));
    }
    
    return numHit;
    
  } // collides()
  
  // move the segments
  public void advance(LinkedList<Sprite> addTheseSprites,
                      LinkedList<Sprite> killTheseSprites,
                      LinkedList<StoryEvent> newStoryEvents) {

    mTime++;

    // analyse chains
    int start = 0;
    int maxBoostTime = 0,
        maxBoostDirec = 0,
        sumBoostTime = 0;
    double sumXVel = 0.0,
           minXPos = Env.screenWidth()/2,
           maxXPos = Env.screenWidth()/2;
    for ( int k = 0 ; k < mSegments.size(); k++ ) {
      
      // collect statistics for this chain
      Segment seg = mSegments.get(k); 
      double xp = seg.xPos() + Segment.width()/2,
             xv = seg.xVel();
      int bt = seg.boostTime();
      if ( bt > maxBoostTime ) {
        maxBoostTime = bt;
        maxBoostDirec = ( (xv > 0) ? +1 : -1 );
      }
      sumBoostTime += bt;
      sumXVel += xv;
      if ( xp < minXPos ) minXPos = xp;
      if ( xp > maxXPos ) maxXPos = xp;
      
      // reached end of chain, now make changes
      if ( k >= mChains.size() || !mChains.get(k) ) {
        int boostTime = sumBoostTime / (k-start+1);
        double xVel = sumXVel / (k-start+1);
        if ( minXPos < 0 || maxXPos >= Env.screenWidth() ) {
          boostTime = maxBoostTime = 0;
          if ( minXPos < 0 && xVel < 0.0 )  xVel = 0.0;
          if ( minXPos >= 0 && xVel > 0.0 ) xVel = 0.0;
        }
        if ( maxBoostTime > 0 ) {
          xVel = maxBoostDirec * Segment.kBoostSpeed;
        }
        for ( int n = start ; n <= k ; n++ ) {
          Segment s = mSegments.get(n);
          int x = Env.screenWidth()/2 - Segment.width()/2
                + (kSegmentSpacing*(2*(n-start) - (k-start)))/2;
          s.setRestPos(x);
          s.setXVel(xVel, boostTime);
        }
        
        // reset statistics
        start = k+1;
        maxBoostTime = maxBoostDirec = sumBoostTime = 0;
        sumXVel = 0.0;
        minXPos = maxXPos = Env.screenWidth()/2;
      }
      
    }
    
    // move the segments
    minXPos = -1.0e10;
    for ( int k = 0 ; k < mSegments.size(); k++ ) {
      Segment seg = mSegments.get(k);
      seg.setTime(mTime + k*kSegmentDelay);
      boolean contact = seg.advance(minXPos);
      if ( contact && k > 0 ) mChains.set(k-1, true);
      minXPos = seg.xPos() + kSegmentSpacing;
    }
    
    // check the mouse
    MouseMonitor.State mouse = Env.mouse().getState();
    int boostDirection = 0;
    if ( mouse.b1 ) {
      if ( mMouseReady ) boostDirection = -1;
      mMouseReady = false;
    } else if ( mouse.b2 ) {
      if ( mMouseReady ) boostDirection = +1;
      mMouseReady = false;
    } else {
      mMouseReady = true;
    }
    if ( !mPlayable ) mMouseReady = false;
    
    // whether the mouse is over a segment
    int hitIndex = -1;
    int smallestDistance = Env.screenWidth();
    for ( int k = mSegments.size()-1 ; k >= 0 ; k-- ) {
      Segment seg = mSegments.get(k);
      int d = seg.mouseDistance(mouse.x, mouse.y);
      if ( d >= 0 && d <= kMouseMargin && d < smallestDistance ) {
        hitIndex = k;
        smallestDistance = d;
      }
    }
    mHighlightIndex = ( mMouseReady ? hitIndex : -1 );
    
    // boost the segment chain
    if ( boostDirection != 0 && hitIndex != -1 ) {
      mSegments.get(hitIndex).boost(boostDirection);
      
      if ( boostDirection < 0 ) {
        for ( int k = hitIndex-1 ; k >= 0 ; k-- ) {
          if ( mChains.get(k) ) mSegments.get(k).boost(boostDirection);
          else break;
        }
        if ( hitIndex < mChains.size() ) mChains.set(hitIndex, false);
      }
      
      if ( boostDirection > 0 ) {
        for ( int k = hitIndex+1 ; k < mSegments.size() ; k++ ) {
          if ( mChains.get(k-1) ) mSegments.get(k).boost(boostDirection);
          else break;
        }
        if ( hitIndex > 0 ) mChains.set(hitIndex-1, false);
      }
    }
    
    // detonate the segments
    if ( mSelfDestructTime > 0 ) {
      if ( mSegments.size() == 0 ) {
        mSelfDestructTime = 0;
      } else if ( --mSelfDestructTime == 0 ) {
        mSelfDestructTime = Math.max(1, kSelfDestructDelay
                                        /(1 + mSegments.size()/8));
        int index = Env.randomInt(mSegments.size());
        int x = (int)Math.round(mSegments.get(index).xPos()) + 2,
            y = (int)Math.round(mSegments.get(index).yPos()) + 1;
        removeSegment(index);
        addTheseSprites.add(new Explosion(x, y, true));
      }
    }
    
    assert( (mChains.size() == mSegments.size()-1) ||
            (mChains.size() == 0 && mSegments.size() <= 1) );
    
  } // Sprite.advance()

  // display
  public void draw(Graphics2D g2) {
    
    for ( int k = mSegments.size()-1 ; k >= 0 ; k-- ) {
      Segment seg = mSegments.get(k);
      boolean highlight = ( k == mHighlightIndex );
      seg.draw(g2, highlight);
    }
    
  } // Sprite.draw()

} // class Chain
