/*
 *  Env.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.util.*;

// global stuff
public class Env {

  static private Component kOwner;

  // enumeration of directions (used in various places)
  static public final int kDirectionUp    = 0,
                          kDirectionDown  = 1,
                          kDirectionLeft  = 2,
                          kDirectionRight = 3;
  
  // size of game screen
  static private final int kScreenWidth  = 600, //512, 
                           kScreenHeight = 384;
  
  // whether to display debug messages, timing statistics
  static private boolean kDebugMode        = false,
                         kPerformanceStats = false;
  
  // whether to use a double-size window for the game
  static private boolean kMagnifyScreenMode = false;
  
  // whether to use full screen mode
  static private boolean kFullScreenMode = false;
  
  // assorted helper objects
  static private Random       kRandom;
  static private KeyMonitor   kKeyMonitor;
  static private MouseMonitor kMouseMonitor;
  static private Resources    kResources;
  //static private Sounds       kSounds;
  
  // some graphics objects
  static private GraphicsEnvironment   kGraphicsEnvironment;
  static private GraphicsDevice        kGraphicsDevice;
  static private GraphicsConfiguration kGraphicsConfiguration;
  
  // this sets up a global Env for the applet
  static public void initialize(Component owner) {
  
    kOwner = owner;

    kRandom       = new Random();
    kKeyMonitor   = new KeyMonitor(owner);
    kMouseMonitor = new MouseMonitor();
    kResources    = new Resources();
    //kSounds       = new Sounds();
    
    kGraphicsEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment();
    kGraphicsDevice = kGraphicsEnvironment.getDefaultScreenDevice();
    kGraphicsConfiguration = owner.getGraphicsConfiguration();
    
  } // initialize()

  // this sets up a global Env for an unspecified owner
  // (some features of the Env will not be accessible)
  static public void initialize() {
  
    kOwner = null;

    kRandom       = new Random();
    kKeyMonitor   = null;
    kMouseMonitor = null;
    kResources    = new Resources();
    //kSounds       = new Sounds();
    
    kGraphicsEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment();
    kGraphicsDevice = kGraphicsEnvironment.getDefaultScreenDevice();
    kGraphicsConfiguration = kGraphicsDevice.getDefaultConfiguration();
    
  } // initialize()

  // whether debug messages, timing statistics should be displayed
  static public boolean debugMode() { return kDebugMode; }
  static public boolean performanceStats() { return kPerformanceStats; }
  static public void setVerbose() { kDebugMode = kPerformanceStats = true; }

  // return dimensions of applet component
  static public int screenWidth()  { return kScreenWidth; }
  static public int screenHeight() { return kScreenHeight; }

  // return reference to keyboard monitor
  static public KeyMonitor keys() { return kKeyMonitor; }
  
  // return reference to mouse monitor
  static public MouseMonitor mouse() { return kMouseMonitor; }
  
  // return reference to game resources
  static public Resources resources() { return kResources; }
  
  // return reference to game audio
  //static public Sounds sounds() { return kSounds; }

  // return a non-transparent buffered image suited to the graphics device
  static public BufferedImage createOpaqueImage(int width, int height) {
    
    assert( kGraphicsConfiguration != null );
    return kGraphicsConfiguration.createCompatibleImage(width, height,
                                                        Transparency.OPAQUE);
    
  } // createOpaqueImage()
  
  // return a transparent buffered image suited to the graphics device
  static public BufferedImage createTranslucentImage(int width, int height) {
    
    assert( kGraphicsConfiguration != null );
    return kGraphicsConfiguration.createCompatibleImage(width, height,
                                                    Transparency.TRANSLUCENT);
    
  } // createTranslucentImage()
  
  // return a bit-masked buffered image suited to the graphics device
  static public BufferedImage createBitmaskImage(int width, int height) {
    
    assert( kGraphicsConfiguration != null );
    return kGraphicsConfiguration.createCompatibleImage(width, height,
                                                    Transparency.BITMASK);
    
  } // createBitmaskImage()
  
  // comment on available accelerated memory
  static public void reportAcceleratedMemory() {

    if ( !kPerformanceStats || kGraphicsDevice == null ) return;
    int am = kGraphicsDevice.getAvailableAcceleratedMemory();
    System.out.println("accelerated memory: "  + (am/1024) + "K");
    
  } // reportAcceleratedMemory();
  
  // magnification of the game screen (bit of a hack)
  static public boolean magnifyScreenMode() { return kMagnifyScreenMode; }
  static public float magnifyScreenFactor() { 
    return ( kMagnifyScreenMode ? 2.0f : 1.0f ); 
  }
  static public boolean magnifyModePossible(int width, int height) {
    return ( width > 2*kScreenWidth+40 && height > 2*kScreenHeight+40 );
  }
  static public void setNormalScreenMode() { kMagnifyScreenMode = false; }
  static public void setMagnifyScreenMode() { kMagnifyScreenMode = true; }
  
  // experiment with running the game in full screen mode
  static public boolean fullScreenMode() { return kFullScreenMode; }
  static public void setFullScreenMode() { kFullScreenMode = true; }
  static public void validateFullScreenMode() {
    assert( kGraphicsDevice != null );
    if ( kFullScreenMode && !kGraphicsDevice.isFullScreenSupported() ) {
      kFullScreenMode = false;
      if ( kDebugMode ) System.out.println("full screen mode not supported");
    }
  }
  static public void startFullScreenMode() {
    assert( kFullScreenMode && kGraphicsDevice != null );
    assert( kOwner instanceof Window );
    kGraphicsDevice.setFullScreenWindow((Window)kOwner);
    changeDisplayMode();
  }
  static public void stopFullScreenMode() {
    if ( !kFullScreenMode || kGraphicsDevice == null ) return;
    Window fsw = kGraphicsDevice.getFullScreenWindow();
    if ( fsw != null ) fsw.dispose();
    kGraphicsDevice.setFullScreenWindow(null);
  }
  static public void changeDisplayMode() {
    if ( !kFullScreenMode || kGraphicsDevice == null ||
         !kGraphicsDevice.isDisplayChangeSupported() ) return;
    try {
      DisplayMode dm = new DisplayMode(512, 384, 32, 
                                       DisplayMode.REFRESH_RATE_UNKNOWN);
      kGraphicsDevice.setDisplayMode(dm);
      return;
    } catch (IllegalArgumentException ex) {}
    try {
      DisplayMode dm = new DisplayMode(640, 480, 32, 
                                       DisplayMode.REFRESH_RATE_UNKNOWN);
      kGraphicsDevice.setDisplayMode(dm);
      return;
    } catch (IllegalArgumentException ex) {}
  }
  
  // called when a break in the action (to collect garbage) won't be noticed
  static public void naturalBreak() { System.gc(); }
  
  // end the game, close the window (could be done better?)
  static public void exitGame() {
    
    if ( kOwner != null && kOwner instanceof MainWindow ) {
      ((MainWindow)kOwner).exit();
    } else {
      System.exit(0);
    }
    
  } // exitGame()
  
  // assorted functions for returning random numbers
  static public float randomFloat() { // in range [0,1]
    return kRandom.nextFloat(); 
  } // randomFloat()
  static public float randomFloat(float a, float b) { // in range [a,b] 
    return ( a + (b-a)*kRandom.nextFloat() ); 
  } // randomFloat()
  static public double randomDouble() { // in range [0,1]
    return kRandom.nextDouble(); 
  } // randomDouble()
  static public double randomDouble(double a, double b) { // in range [a,b] 
    return ( a + (b-a)*kRandom.nextDouble() ); 
  } // randomDouble()
  static public int randomInt(int n) { // in range [0,n-1] 
    return kRandom.nextInt(n); 
  } // randomInt()
  static public int randomInt(int a, int b) { // in range [a,b]
    if ( a > b )      return ( b + kRandom.nextInt(a-b+1) );
    else if ( a < b ) return ( a + kRandom.nextInt(b-a+1) );
    else              return a;
  } // randomInt()
  static public boolean randomBoolean() { 
    return kRandom.nextBoolean(); 
  } // randomBoolean()

  // assorted modulo-type functions
  static public int fold(int a, int b) {
    // result is between 0 and (b-1)
    if ( a >= 0 ) return (a%b);
    else {
      int temp = b + (a%b);
      return ( (temp==b) ? 0 : temp );
    }
  } // fold()
  static public double fold(double a, double b) {
    // result is in interval [0,b)
    // (probably a more efficient way of doing this?)
    return ( a - b*Math.floor(a/b) );
  } // fold()
  static public float fold(float a, float b) {
    // result is in interval [0,b)
    // (probably a more efficient way of doing this?)
    return ( a - b*(float)Math.floor(a/b) );
  } // fold()
  
  // more modulo-type functions
  static public float foldNearTo(float a, float target, float modSize) {
    return ( target + fold(a-target+0.5f*modSize, modSize) - 0.5f*modSize );
  } // foldNearTo()
  
} // class Env
