/*
 *  Explosion.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.image.*;
import java.util.LinkedList;

// a bomb exploding
public class Explosion extends Sprite {

  // how sprite is drawn relative to other sprites
  static private final int kScreenLayer = 70;

  // image details
  static private final int kNumImages = 8;
  static private BufferedImage kImages[] = null;
  static private final int kImagePos[][] = { {   8, 176,  31,  26 },
                                             {  47, 176,  32,  27 },
                                             {  90, 170,  36,  35 },
                                             { 139, 136,  98,  95 },
                                             { 247, 122, 122, 122 },
                                             {  59, 242, 107, 108 },
                                             { 178, 264,  59,  60 },
                                             { 245, 278,  35,  33 } };
  static private final int kImageOffset[][] = { { 36, 45 }, 
                                                { 36, 45 }, 
                                                { 34, 42 }, 
                                                {  0, 12 }, 
                                                {  0,  0 },
                                                { 10,  8 }, 
                                                { 34, 31 }, 
                                                { 35, 45 } };
  
  // relative to incoming position
  static private final int kShiftX = -36-7,
                           kShiftY = -45-4;
  
  // speed of animation
  static private final int kFrameTimes[] = { 1, 1, 1, 2, 5, 2, 1, 1 };
  
  // effective size of blast
  static private final double kDeadlyRadius = 0.75;
  
  // position of centre of explosion
  private int mXPos,
              mYPos;
  
  // animation frame currently showing
  private int mFrame;
  
  // timer for animation
  private int mTime;
  
  // true if the explosion causes no damage
  private boolean mIsHarmless;
  
  // true when the explosion takes effect
  private boolean mDeadly;
  
  // number of segments hit by the explosion (including this one)
  private int mNumKilled;
  
  // keep track of the segment chain
  private Chain mChain;
  
  // load resources
  static void initialize() {
    
    if ( kImages != null ) return;

    kImages = new BufferedImage[kNumImages];
    for ( int k = 0 ; k < kNumImages ; k++ ) {
      kImages[k] = Env.resources().getSprite(Resources.kSprites, 
                                             kImagePos[k][0], 
                                             kImagePos[k][1], 
                                             kImagePos[k][2], 
                                             kImagePos[k][3],
                                             false, false);
    }
    
  } // initialize()
  
  // constructor
  public Explosion(int x, int y, boolean isHarmless) {
  
    super(kScreenLayer);
    initialize();
    
    mXPos = x;
    mYPos = y;
    
    mIsHarmless = isHarmless;
    
    mFrame = 0;
    mTime  = 0;
    
    mDeadly = false;
    mNumKilled = 0;
    
  } // constructor
  
  // keep track of the chain
  public void observeArrival(Sprite newSprite) { 
    
    if ( newSprite instanceof Chain ) {
      assert( mChain == null );
      mChain = (Chain)newSprite;
    }
    
  } // Sprite.observeArrival()
  
  // keep track of the chain
  public void observeDeparture(Sprite deadSprite) {

    if ( deadSprite instanceof Chain ) {
      assert( mChain == (Chain)deadSprite );
      mChain = null;
    }
  
  } // observeDeparture()

  // progress the animation
  public void advance(LinkedList<Sprite> addTheseSprites,
                      LinkedList<Sprite> killTheseSprites,
                      LinkedList<StoryEvent> newStoryEvents) {
    
    mTime++;
    
    int sum = 0;
    for ( mFrame = 0 ; mFrame < kNumImages ; mFrame++ ) {
      sum += kFrameTimes[mFrame];
      if ( mTime <= sum ) break;
    }
    mDeadly = ( mFrame == 4 && mTime == sum-kFrameTimes[mFrame]+1 );
    if ( mFrame >= kNumImages-2 ) killTheseSprites.add(this);
    
  } // Sprite.advance()

  // check for explosion destroying chain segments
  public void interact() { 
    
    if ( !mDeadly ) return;
    
    assert( mChain != null );
    assert( mNumKilled == 0 );

    int x = mXPos + kImageOffset[4][0] + kShiftX + kImagePos[4][2]/2,
        y = mYPos + kImageOffset[4][1] + kShiftY + kImagePos[4][3]/2,
        r = (int)Math.round(kDeadlyRadius*kImagePos[4][2]/2);
    mNumKilled = mChain.collides(x, y, r, !mIsHarmless) + 1;
    
  } // Sprite.interact()

  // count the dead
  public void aftermath(LinkedList<Sprite>     addTheseSprites, 
                        LinkedList<Sprite>     killTheseSprites,
                        LinkedList<StoryEvent> newStoryEvents) { 

    if ( mNumKilled > 0 ) {
      newStoryEvents.add(new CentipedeStory.EventExplosion(mNumKilled));
      mNumKilled = -1;
    }
    
  } // Sprite.aftermath
  
  // draw the explosion
  public void draw(Graphics2D g2) {
    
    g2.drawImage(kImages[mFrame], 
                 mXPos + kImageOffset[mFrame][0] + kShiftX, 
                 mYPos + kImageOffset[mFrame][1] + kShiftY, 
                 null);

  } // Sprite.draw()

  // check the images
  private void drawTest(Graphics2D g2) {
    
    AffineTransform oldTransform = g2.getTransform();
    AffineTransform at = AffineTransform.getScaleInstance(3,3);
    g2.setTransform(at);
    
    g2.setColor(Color.MAGENTA);

    int x = kImageOffset[mFrame][0],
        y = kImageOffset[mFrame][1];
    int w = kImagePos[mFrame][2],
        h = kImagePos[mFrame][3];
    g2.fillRect(x, y, w, h);
    g2.drawImage(kImages[mFrame], x, y, null);

    g2.setTransform(oldTransform);
    
  } // drawTest()
  
} // class Explosion
