/*
 *  Flower.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.LinkedList;

// a doomed plant
public class Flower extends Sprite {

  // how sprite is drawn relative to other sprites
  static private final int kScreenLayer = 20;

  // flower stem and centre
  static private BufferedImage kStemImage   = null,
                               kCentreImage = null;
  static private final int kStemHeight      = 80,
                           kStemWidth       = 4,
                           kStemYPos        = Background.groundLevel(),
                           kCentreRadius    = 7,
                           kCentreYPos      = kStemYPos - kStemHeight;
  static private final Color kStemColour    = new Color(0.0f, 0.7f, 0.0f),
                             kCentreColour  = new Color(1.0f, 1.0f, 0.2f);
  
  // flower petals
  static private final int kNumPetals = 10;
  static private final Color kPetalColours[] = { new Color(0.4f,0.5f,1.0f),
                                                 new Color(1.0f,0.5f,1.0f),
                                                 new Color(1.0f,0.5f,0.0f),
                                                 new Color(0.7f,0.4f,1.0f),
                                                 new Color(1.0f,0.2f,0.2f) };
  
  // drawing details
  static private final Color kBlankColour = new Color(0, 0, 0, 0),
                             kLineColour  = new Color(0.0f, 0.0f, 0.0f);
  static private final BasicStroke kLineStroke = new BasicStroke(0.5f);
  
  // movement
  static private final int    kStartXPos  = Env.screenWidth() + 40,
                              kSpeed      = -1,
                              kDetachXPos = Env.screenWidth()/2;
  static private final double kGravity    = 0.21;
  
  // current position
  private int mXPos;
  
  // current position of centre
  private double mCentreYPos,
                 mCentreYVel;
  
  // separate petals
  private Petal mPetals[];
  
  // whether the petals have been launched yet
  private boolean mPetalsDetached;
  
  // prepare images
  static public void initialize() {
    
    if ( kStemImage   == null ) prepareStem();
    if ( kCentreImage == null ) prepareCentre();
    
    Petal.initialize();
    
  } // initialize()
  
  // draw stem image
  static private void prepareStem() {
    
    int margin = 2;
    int width  = kStemWidth + 2*margin,
        height = kStemHeight + 2*margin;
    kStemImage = Env.createTranslucentImage(width, height);
    
    Graphics2D g2 = kStemImage.createGraphics();
    g2.setBackground(new Color(0,0,0,0));
    g2.clearRect(0, 0, width, height);
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
                        RenderingHints.VALUE_ANTIALIAS_ON);
    
    int x0 = (width - kStemWidth)/2,
        y0 = margin + kCentreRadius;
    g2.setColor(kStemColour);
    g2.fillRect(x0, y0, kStemWidth, kStemHeight);
    g2.setColor(kLineColour);
    g2.setStroke(kLineStroke);
    g2.drawLine(x0, y0, x0, y0+kStemHeight);
    g2.drawLine(x0+kStemWidth, y0, x0+kStemWidth, y0+kStemHeight);

    g2.dispose();
    
  } // prepareStem()
  
  // draw stem image
  static private void prepareCentre() {
    
    int margin = 2;
    int width  = 2*kCentreRadius + 2*margin,
        height = 2*kCentreRadius + 2*margin;
    kCentreImage = Env.createTranslucentImage(width, height);
    
    Graphics2D g2 = kCentreImage.createGraphics();
    g2.setBackground(new Color(0,0,0,0));
    g2.clearRect(0, 0, width, height);
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
                        RenderingHints.VALUE_ANTIALIAS_ON);
    
    int x = width/2,
        y = height/2;
    Ellipse2D centre = new Ellipse2D.Double(x-kCentreRadius, 
                                            y-kCentreRadius,
                                            2*kCentreRadius, 
                                            2*kCentreRadius);
    g2.setColor(kCentreColour);
    g2.fill(centre);
    g2.setColor(kLineColour);
    g2.setStroke(kLineStroke);
    g2.draw(centre);
    
    g2.dispose();
    
  } // prepareCentre()

  // constructor
  public Flower(int level) {
    
    super(kScreenLayer);
    initialize();
    
    mXPos = kStartXPos;
    
    mCentreYPos = kCentreYPos - kCentreImage.getHeight()/2;
    mCentreYVel = 0.0;
    
    mPetals = new Petal[kNumPetals];
    Color col = kPetalColours[(level % kPetalColours.length)];
    double dTheta = 2.0*Math.PI/kNumPetals;
    for ( int k = 0 ; k < kNumPetals ; k++ ) {
      int i = 2*k + ((k >= kNumPetals/2) ? 1 : 0);
      double theta = (i+0.3)*dTheta;
      mPetals[k] = new Petal(mXPos, kCentreYPos, theta, col);
    }
    
    mPetalsDetached = false;
    
  } // constructor
  
  // ticks until the petals fall off (may be negative)
  public int timeTillDetach() { return ((kDetachXPos - mXPos)/kSpeed); } 
  
  // move
  public void advance(LinkedList<Sprite> addTheseSprites,
                      LinkedList<Sprite> killTheseSprites,
                      LinkedList<StoryEvent> newStoryEvents) {

    mXPos += kSpeed;
    
    boolean detachPetals = ( mXPos == kDetachXPos );
    for ( int k = 0 ; k < kNumPetals ; k++ ) {
      if ( detachPetals ) mPetals[k].detach();
      mPetals[k].advance(kSpeed);
    }
    
    if ( detachPetals ) mPetalsDetached = true;
    if ( mPetalsDetached ) {
      mCentreYPos += mCentreYVel;
      mCentreYVel += kGravity;
    }
    
    if ( mXPos < -kCentreRadius ) killTheseSprites.add(this);
    
  } // Sprite.advance()

  // display
  public void draw(Graphics2D g2) {

    if ( !mPetalsDetached ) {
      g2.drawImage(kStemImage, 
                   mXPos-kStemImage.getWidth()/2,
                   kStemYPos-kStemImage.getHeight(), 
                   null);
    }
    
    for ( int k = 0 ; k < kNumPetals ; k++ ) {
      mPetals[k].draw(g2);
    }

    g2.drawImage(kCentreImage, 
                 mXPos-kCentreImage.getWidth()/2,
                 (int)Math.round(mCentreYPos), 
                 null);
    
  } // Sprite.draw()

} // class Flower
