/*
 *  Incoming.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.util.LinkedList;

// a falling bomb
public class Incoming extends Sprite {

  // how sprite is drawn relative to other sprites
  static private final int kScreenLayer = 60;

  // flying segment image
  static private final int kFlyingImageX      = 127,
                           kFlyingImageY      = 85,
                           kFlyingImageWidth  = 28,
                           kFlyingImageHeight = 28,
                           kFlyingImageBody   = kFlyingImageHeight - 9;
  static private BufferedImage kFlyingImage = null;
  
  // standing segment image
  static private final int kStandingImageX      = 54,
                           kStandingImageY      = 35,
                           kStandingImageWidth  = 20,
                           kStandingImageHeight = 26;
  static private final int kStandingDeltaX      = 0,
                           kStandingDeltaY      = 0;
  static private BufferedImage kStandingImage = null;
  
  // collision region
  static private final int kCollisionX = 10,
                           kCollisionY = 11,
                           kCollisionR = 8;

  // projectile parameters
  static private final int    kFlightYTop   = kFlyingImageHeight,
                              kFlightYStart = Env.screenHeight()/2,
                              kFlightYEnd   = Background.groundLevel()
                                              - kFlyingImageBody,
                              kDropTime     = 60,
                              kDropDistance = kFlightYEnd - kFlightYTop;
  static private final double kParabCoeff   = kDropDistance
                                              /(double)(kDropTime*kDropTime);
  static private final int    kUpDistance   = kFlightYStart - kFlightYTop,
                              kStartTime    = -(int)Math.round(
                                                  Math.sqrt(kUpDistance
                                                            /kParabCoeff));
  
  // bounce parameters
  static private final int kBounceHeight = 10,
                           kBounceYTop   = kFlightYEnd - kBounceHeight,
                           kBounceTime   = 2*(int)Math.round(
                                                    Math.sqrt(kBounceHeight
                                                              /kParabCoeff)),
                           kStandTime    = kBounceTime/2 - 5,
                           kStandYPos    = Background.groundLevel()
                                           - kStandingImageHeight;
  
  // standing means moving at background speed
  static private final double kStandingSpeed = -1.0,
                              kStandingDecel = 0.2;
  static private final int    kMoveTime      = 10;
  
  // position, etc.
  private double mXPos,
                 mYPos;
  private double mXVel;
  
  // counter for flight
  private int mTime;
  
  // flying or standing
  private boolean mStanding;
  
  // ready to join the chain
  private boolean mAddToChain;
  
  // cue detonation
  private boolean mExplode;
  
  // keep track of the chain
  private Chain mChain = null;
  
  // prepare resources
  static public void initialize() {
    
    if ( kFlyingImage != null ) return;
    
    kFlyingImage = Env.resources().getSprite(Resources.kSprites, 
                                             kFlyingImageX, 
                                             kFlyingImageY, 
                                             kFlyingImageWidth, 
                                             kFlyingImageHeight,
                                             false, false);
    kStandingImage = Env.resources().getSprite(Resources.kSprites, 
                                               kStandingImageX, 
                                               kStandingImageY, 
                                               kStandingImageWidth, 
                                               kStandingImageHeight,
                                               false, false);

  } // initialize()
  
  // constructor
  public Incoming(int xTarget) {
    
    super(kScreenLayer);
    initialize();
    
    mTime = kStartTime;
    mYPos = calculateYPos(mTime);
    
    boolean fromRight = Env.randomBoolean();
    if      ( xTarget < Env.screenWidth()/3 )     fromRight = false;
    else if ( xTarget > (2*Env.screenWidth())/3 ) fromRight = true;
    
    mXPos = ( fromRight ? Env.screenWidth() : -kFlyingImageWidth );
    double deltaT = kDropTime - kStartTime;
    double deltaX = (xTarget-kFlyingImageWidth/2) - mXPos;
    mXVel = deltaX/deltaT;
    
    mStanding = false;
    mAddToChain = false;
    mExplode = false;
    
  } // constructor
  
  // keep track of the chain
  public void observeArrival(Sprite newSprite) { 
    
    if ( newSprite instanceof Chain ) {
      assert( mChain == null );
      mChain = (Chain)newSprite;
    }
    
  } // Sprite.observeArrival()
  
  // keep track of the chain
  public void observeDeparture(Sprite deadSprite) {

    if ( deadSprite instanceof Chain ) {
      assert( mChain == (Chain)deadSprite );
      mChain = null;
    }
  
  } // observeDeparture()

  // evaluate parabolic trajectory
  static private double calculateYPos(int t) { 
    
    // main parabola
    if ( t <= kDropTime ) return (kFlightYTop + kParabCoeff*t*t); 
    
    // bounce
    if ( t <= kDropTime + kStandTime ) {
      t = t - kDropTime - kBounceTime/2;
      double y = kBounceYTop + kParabCoeff*t*t;
      return Math.min(y, kFlightYEnd);
    }
    
    // just standing
    return kStandYPos;
    
  } // calculateYPos
  
  // move the bomb
  public void advance(LinkedList<Sprite> addTheseSprites,
                      LinkedList<Sprite> killTheseSprites,
                      LinkedList<StoryEvent> newStoryEvents) {
  
    mTime++;
    if ( mTime == kDropTime ) {
      mXVel = 0.5*(mXVel + kStandingSpeed);
    }
    if ( mTime == kDropTime+1 ) {
      newStoryEvents.add(new CentipedeStory.EventLanding());
    }
    if ( mTime == kDropTime + kStandTime ) {
      mXPos += kStandingDeltaX;
      mStanding = true;
    }
    if ( mTime == kDropTime + kStandTime + kMoveTime ) {
      mAddToChain = true;
    }
    
    if ( mStanding ) {
      double dv = mXVel - kStandingSpeed;
      if      ( dv > +kStandingDecel ) mXVel -= kStandingDecel;
      else if ( dv < -kStandingDecel ) mXVel += kStandingDecel;
      else                             mXVel = kStandingSpeed; 
    }
    
    mYPos = calculateYPos(mTime);
    mXPos += mXVel;
    
  } // Sprite.advance()

  // check for collision with chain segments
  public void interact() { 
    
    assert( mChain != null );
    
    int numHit = mChain.collides(mXPos + kCollisionX, 
                                 mYPos + kCollisionY, 
                                 kCollisionR, false);
    if ( numHit > 0 ) {
      if ( mTime <= kDropTime ) mExplode    = true;
      else                      mAddToChain = true;
    }
    
  } // Sprite.interact()
  
  // add this bomb to the chain
  public void aftermath(LinkedList<Sprite>     addTheseSprites, 
                        LinkedList<Sprite>     killTheseSprites,
                        LinkedList<StoryEvent> newStoryEvents) { 
    
    if ( mExplode ) {
      addTheseSprites.add(new Explosion((int)Math.round(mXPos), 
                                        (int)Math.round(mYPos), false));
      killTheseSprites.add(this);
      return;
    }
    
    if ( mAddToChain ) {
      assert( mChain != null );
      mChain.addSegment((int)Math.round(mXPos), mXVel);
      killTheseSprites.add(this);
      return;
    }
    
  } // Sprite.aftermath()

  // display the sprite
  public void draw(Graphics2D g2) {
  
    int x = (int)Math.round(mXPos),
        y = (int)Math.round(mYPos);
    g2.drawImage((mStanding ? kStandingImage : kFlyingImage), 
                 x, y, null);

  } // Sprite.draw()

  // test sprite images
  private void drawTest(Graphics2D g2) {
    
    AffineTransform oldTransform = g2.getTransform();
    AffineTransform at = AffineTransform.getScaleInstance(4,4);
    g2.setTransform(at);
    
    g2.setColor(Color.MAGENTA);
    g2.fillRect(0, 0, kFlyingImageWidth, kFlyingImageHeight);
    g2.drawImage(kFlyingImage, 0, 0, null);

    g2.fillRect(0, 30, kStandingImageWidth, kStandingImageHeight);
    g2.drawImage(kStandingImage, 0, 30, null);
    
    g2.setTransform(oldTransform);
    
  } // drawTest()
  
} // class Incoming
