/*
 *  MainWindow.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.awt.image.BufferStrategy;
import javax.swing.JFrame;

// the main game window
public class MainWindow extends JFrame implements Runnable {

  // main method
  public static void main(String args[]) {

    int drawMode = 0;
    
    //!!!
    //Env.setVerbose();
    //!!!
    
    boolean unknownArg = false;
    for ( int k = 0 ; k < args.length ; k++ ) {
      String arg = args[k].toLowerCase();
      if      ( arg.equals("debug") )       Env.setVerbose();
      else if ( arg.equals("fullscreen") )  Env.setFullScreenMode();
      else if ( arg.equals("use_dd") )      drawMode = 1;
      else if ( arg.equals("use_ogl") )     drawMode = 2;
      else                                  unknownArg = true;
    }
    if ( unknownArg ) {
      System.out.println("java -jar <file>.jar [options]");
      System.out.println("options: debug, fullscreen, use_dd, use_ogl");
      System.exit(0);
    }
    
    if ( drawMode == 1 ) {
      System.setProperty("sun.java2d.ddforcevram","true");
      System.setProperty("sun.java2d.translaccel","true");
    } else if ( drawMode == 2 ) {
      System.setProperty("sun.java2d.opengl","true");
    }

    //if ( Env.debugMode() ) {
    //  System.setProperty("sun.java2d.trace","count");
    //}
    
    new MainWindow();

  } // main()

  // this object collects timing statistics as the game is running
  private class TimingStats {
    private int  mNumTicks, mNumOverruns, mNumDrawSkips;
    private long mNumNanos, mNumNanosInAdvance, mNumNanosInDraw;
    private long mPeakNanos, mPeakNanosInAdvance, mPeakNanosInDraw;
    public TimingStats() { 
      clear(); 
    }
    public void clear() {
      mNumTicks = mNumOverruns = mNumDrawSkips = 0;
      mNumNanos = mNumNanosInAdvance = mNumNanosInDraw = 0;
      mPeakNanos = mPeakNanosInAdvance = mPeakNanosInDraw = 0;
    }
    public void update(long    nanosInTick,
                       long    nanosInAdvance,
                       long    nanosInDraw,
                       boolean tickHasOverrun,
                       boolean tickSkippedDraw) {
      mNumTicks++;
      if ( tickHasOverrun ) mNumOverruns++;
      if ( tickSkippedDraw ) mNumDrawSkips++;
      mNumNanos += nanosInTick;
      mPeakNanos = Math.max(nanosInTick, mPeakNanos);
      mNumNanosInAdvance += nanosInAdvance;
      mPeakNanosInAdvance = Math.max(nanosInAdvance, mPeakNanosInAdvance);
      mNumNanosInDraw += nanosInDraw;
      mPeakNanosInDraw = Math.max(nanosInDraw, mPeakNanosInDraw);
    }
    public String toString() {
      int  numTicks = Math.max(1, mNumTicks);
      long numNanos = Math.max(1, mNumNanos);
      return new String(mNumTicks
                        + " frames, "
                        + String.format("%.1f", mNumTicks/(numNanos*1.0e-9f))
                        + " per sec (max "
                        + String.format("%.1f", mPeakNanos*1.0e-6f)
                        + "ms), "
                        + (100*mNumNanosInAdvance)/(numTicks*kNanosPerTick)
                        + "% in advance (peak "
                        + (100*mPeakNanosInAdvance)/kNanosPerTick
                        + "%), "
                        + (100*mNumNanosInDraw)/(numTicks*kNanosPerTick)
                        + "% in draw (peak "
                        + (100*mPeakNanosInDraw)/kNanosPerTick
                        + "%), "
                        + mNumOverruns 
                        + " overrun (" 
                        + (100*mNumOverruns)/numTicks
                        + "%), "
                        + mNumDrawSkips
                        + " no draw ("
                        + (100*mNumDrawSkips)/numTicks
                        + "%)");
    }
  } // class TimingStats
  
  // desired frame rate
  static private final int  kTicksPerSecond = 30;
  static private final long kNanosPerTick   = Math.round(1.0e9f
                                                         /kTicksPerSecond);
  
  // aim to even out the time-per-tick over this number of ticks
  static private final int kTickHistoryLength = 4;
  
  // game loop must yield to other threads eventually
  static private final int kMaxTicksWithoutYield = 10;
  
  // number of seconds between progress reports on the console (0 to disable)
  static private final int kReportProgressInterval = 1;
  
  // don't collect stats until performance has settled down
  static private final int kNoStatsForEarlyTicks = 5*30;
  
  // if the game runs slow then skip draw updates (in a random pattern)
  static private final int   kMinTicksWithDraw   = 1,
                             kMaxTicksWithDraw   = 3;
  static private final float kAcceptableSlowdown = 1.1f;
  
  private Toolkit        mToolkit        = null;
  private Thread         mMainLoop       = null; 
  private Canvas         mGameCanvas     = null;
  private BufferStrategy mBufferStrategy = null;
  private GameManager    mGameManager    = null;

  // record the times (nanos) of previous ticks
  private long[] mTickHistory      = null;
  private int    mTickHistoryIndex = 0;
  
  // timing statistics
  private TimingStats mTotalStats = null,
                      mLocalStats = null;
  
  // number of times we have skimped on yield
  private int mNumTicksWithoutYield  = 0;
  
  // number of frames before considering skipping a redraw
  private int mNumDrawsBeforeSkip = 0;
  
  // constructor
  public MainWindow() {

    Env.initialize(this);
    Env.reportAcceleratedMemory();
    Env.validateFullScreenMode();
    
    mToolkit = Toolkit.getDefaultToolkit();

    if ( Env.fullScreenMode() ) {
      Env.keys().monitor(this);
      
      setUndecorated(true);
      setIgnoreRepaint(true);
      setResizable(false);
      
      setCursor(mToolkit.createCustomCursor(
                          new BufferedImage(1, 1, BufferedImage.TYPE_INT_ARGB), 
                          new Point(0,0), ""));
      
      Env.startFullScreenMode();
    } else {
      mGameCanvas = new Canvas();
      Dimension canvasDim = new Dimension(Env.screenWidth(), 
                                          Env.screenHeight());
      mGameCanvas.setSize(canvasDim);
      mGameCanvas.setPreferredSize(canvasDim);
      mGameCanvas.setMinimumSize(canvasDim);
      mGameCanvas.setMaximumSize(canvasDim);
      
      Env.keys().monitor(mGameCanvas);
      Env.mouse().monitor(mGameCanvas);
    
      getContentPane().removeAll();
      getContentPane().setLayout(new BorderLayout());
      getContentPane().add(mGameCanvas, BorderLayout.CENTER);

      Dimension screenDim = mToolkit.getScreenSize();
      setLocation(screenDim.width/2 - Env.screenWidth()/2, 
                  screenDim.height/2 - Env.screenHeight()/2);
      setResizable(Env.magnifyModePossible(screenDim.width, screenDim.height));

      setIgnoreRepaint(true);
      pack();   

      if ( isResizable() ) {
        // resizing the window toggles magnification mode (bit messy)
        addComponentListener( 
          new ComponentAdapter() {
            private boolean mIgnoreNextResize = false;
            public void componentResized(ComponentEvent e) { 
              if ( mIgnoreNextResize ) {
                mIgnoreNextResize = false;
                return;
              }
              if ( Env.magnifyScreenMode() ) Env.setNormalScreenMode();
              else                           Env.setMagnifyScreenMode();
              float magFac = Env.magnifyScreenFactor();
              Dimension newDim = new Dimension(
                                       Math.round(magFac*Env.screenWidth()), 
                                       Math.round(magFac*Env.screenHeight()) );
              mGameCanvas.setSize(newDim);
              mGameCanvas.setPreferredSize(newDim);
              mGameCanvas.setMinimumSize(newDim);
              mGameCanvas.setMaximumSize(newDim);
              Dimension screenDim = mToolkit.getScreenSize();
              setLocation(screenDim.width/2 - newDim.width/2, 
                          screenDim.height/2 - newDim.height/2);
              mIgnoreNextResize = true;
              pack();
            }
          }
        );
      }
    }
    
    addWindowListener(
      new WindowAdapter() {
        public void windowClosing(WindowEvent e) { exit(); }
        public void windowDeiconified(WindowEvent e) { start(); }
        public void windowIconified(WindowEvent e) { stop(); }
      }
    );

    addMouseListener(
      new MouseAdapter() {
        public void mouseClicked(MouseEvent e) { requestFocusInWindow(); }
      }
    );
    
    if ( mGameCanvas != null ) mGameCanvas.addMouseListener(
      new MouseAdapter() {
        public void mouseClicked(MouseEvent e) { requestFocusInWindow(); }
      }
    );
    
    requestFocusInWindow();
    validate();
    setVisible(true);

    mTotalStats = new TimingStats();
    mLocalStats = new TimingStats();

    mTickHistory      = new long[kTickHistoryLength];
    mTickHistoryIndex = 0;
    
    if ( Env.fullScreenMode() ) {
      createBufferStrategy(2);
      mBufferStrategy = getBufferStrategy();
    } else {
      mGameCanvas.createBufferStrategy(2);
      mBufferStrategy = mGameCanvas.getBufferStrategy();
    }

    mGameManager = new GameManager(new CentipedeStory());
    setTitle("End of Daisies");

    start();
        
  } // constructor
  
  // start the game loop in a new thread
  public void start() {

    if ( mMainLoop == null ) {
      mMainLoop = new Thread(this);
      mMainLoop.start();
    }
    requestFocusInWindow();

  } // start()
  
  // interrupt the game thread
  public void stop() {
    
    if ( mMainLoop != null ) {
      mMainLoop.interrupt();
    }
    mMainLoop = null;

    Env.keys().reset();
    //Storage.flushCache();
    
    if ( Env.performanceStats() ) {
      System.out.println("Overall: " + mTotalStats.toString());
    }
    
  } // stop()

  // end the game
  public void exit() {
 
    if ( Env.fullScreenMode() ) Env.stopFullScreenMode();
    stop();
    System.exit(0);
    
  } // exit()
  
  // the game loop (mostly concerned with counting nanoseconds)
  public void run() {

    requestFocus();

    mNumTicksWithoutYield  = 0;
    mNumDrawsBeforeSkip = 0;
    
    long prevTickNanos = System.nanoTime();
    prepareTickHistory(prevTickNanos);

    long nextUpdateTime = prevTickNanos + kReportProgressInterval*1000000000;
    mLocalStats.clear();

    int skipEarlyStats = kNoStatsForEarlyTicks;
    
    while ( mMainLoop == Thread.currentThread() ) {
      
      // advance
      mGameManager.advance();
      long nanosAfterAdvance = System.nanoTime();

      // draw
      boolean tickSkipsDraw = gameRunningSlow();
      if ( !tickSkipsDraw ) drawGameScreen();
      long nanosAfterDraw = System.nanoTime();
      
      // wait
      boolean tickHasOverrun = delayUntil(nextTickFromHistory());
      long newTickNanos = System.nanoTime();

      // collect statistics
      updateTickHistory(newTickNanos);
      if ( skipEarlyStats > 0 ) {
        skipEarlyStats--;
      } else {
        updateProgress(newTickNanos - prevTickNanos,
                       nanosAfterAdvance - prevTickNanos,
                       nanosAfterDraw - nanosAfterAdvance,
                       tickHasOverrun, tickSkipsDraw);
        if ( Env.performanceStats() &&
             kReportProgressInterval >= 0 && newTickNanos > nextUpdateTime ) {
          System.out.println(mLocalStats.toString());
          mLocalStats.clear();
          //Env.reportAcceleratedMemory();
          nextUpdateTime = newTickNanos + kReportProgressInterval*1000000000;
        }
      }
      prevTickNanos = newTickNanos;
      
    }
    
    mMainLoop = null;
    
  } // Runnable.run()
  
  // draw the game image onto the off-screen buffer, then flip to the screen
  private void drawGameScreen() {
    
    Graphics g = mBufferStrategy.getDrawGraphics();
    assert( g instanceof Graphics2D );
    
    if ( Env.fullScreenMode() ) drawScreenBorder((Graphics2D)g);
    
    mGameManager.draw((Graphics2D)g);

    g.dispose();
    mToolkit.sync();
    if ( !mBufferStrategy.contentsLost() ) mBufferStrategy.show();
    
  } // drawGameScreen()

  // only use part of the screen in full screen mode, decorate the rest
  // (full screen is just an experiment, so nothing serious here)
  private void drawScreenBorder(Graphics2D g2) {
    
    assert( Env.fullScreenMode() );
    
    Rectangle fullScreen = getBounds();
    if ( fullScreen.width <= Env.screenWidth() &&
         fullScreen.height <= Env.screenHeight() ) return;
    
    g2.setColor(Color.BLACK);
    g2.fill(fullScreen);
    
    int dx = (fullScreen.width - Env.screenWidth())/2,
        dy = (fullScreen.height - Env.screenHeight())/2;
    g2.translate(dx, dy);
    
    int edge = 2;
    g2.setColor(Color.GRAY);
    g2.drawRect(-edge, -edge, 
                Env.screenWidth()+2*edge-1, 
                Env.screenHeight()+2*edge-1);
    
    g2.setClip(0, 0, Env.screenWidth(), Env.screenHeight());
    
  } // drawScreenBorder()

  // wait until the system time has passed the target time
  // and yield to other threads when an opportunity arises
  // (returns true if the target time has passed already)
  // (note: actually sleep(1) instead of yield(), said to be more reliable)
  private boolean delayUntil(long targetNanos) {
    
    if ( targetNanos - System.nanoTime() < 1000 ) {

      if ( mNumTicksWithoutYield < kMaxTicksWithoutYield ) {
        mNumTicksWithoutYield++;
      } else {
        mNumTicksWithoutYield = 0;
        try { Thread.sleep(1); } catch (InterruptedException ex) {}
        //Thread.yield();
      }
      return true;
      
    } else {
    
      mNumTicksWithoutYield = 0;
      try { Thread.sleep(1); } catch (InterruptedException ex) {}
      //Thread.yield();
      while ( targetNanos - System.nanoTime() > 0 ) {}
      return false;
      
    }
    
  } // delayUntil()

  // define an ideal history of tick times based on the current time
  private void prepareTickHistory(long currentNanos) {
    
    for ( int k = 0 ; k < kTickHistoryLength ; k++ ) {
      mTickHistory[k] = currentNanos - (k+1-kTickHistoryLength)*kNanosPerTick;
    }
    mTickHistoryIndex = kTickHistoryLength - 1;
    
  } // prepareTickHistory()
  
  // nominate a time that this tick should end at
  // (try to balance out variations over the history of recorded ticks)
  private long nextTickFromHistory() {
    
    long soonestNanos = mTickHistory[mTickHistoryIndex] + kNanosPerTick;
    for ( int k = 1 ; k < kTickHistoryLength ; k++ ) {
      long t =  mTickHistory[(mTickHistoryIndex+k)%kTickHistoryLength] 
             + (kTickHistoryLength+1-k)*kNanosPerTick;
      if ( t < soonestNanos ) soonestNanos = t;
    }
    return soonestNanos;
    
  } // nextTickFromHistory()
  
  // append the current time to the tick history (a circular array)
  private void updateTickHistory(long currentNanos) {
    
    if ( ++mTickHistoryIndex == kTickHistoryLength ) mTickHistoryIndex = 0;
    mTickHistory[mTickHistoryIndex] = currentNanos;
    
  } // updateTickHistory()
  
  // check whether frame rate is too low
  private boolean gameRunningSlow() {
    
    if ( mNumDrawsBeforeSkip > 0 ) {
      mNumDrawsBeforeSkip--;
      return false;
    }
    
    long meanTick = ( mTickHistory[mTickHistoryIndex]
                    - mTickHistory[(mTickHistoryIndex+1)%kTickHistoryLength] )
                  / (kTickHistoryLength-1);
    if ( meanTick <= kNanosPerTick*kAcceptableSlowdown ) return false;
    
    mNumDrawsBeforeSkip = Env.randomInt(kMinTicksWithDraw, kMaxTicksWithDraw);
    return true;
    
  } // gameRunningSlow()
  
  // update the running statistics and the overall statistics
  private void updateProgress(long    nanosInTick,
                              long    nanosInAdvance,
                              long    nanosInDraw,
                              boolean tickHasOverrun,
                              boolean tickSkipsDraw) {
    
    mTotalStats.update(nanosInTick, nanosInAdvance, 
                       nanosInDraw, tickHasOverrun, tickSkipsDraw);
    mLocalStats.update(nanosInTick, nanosInAdvance, 
                       nanosInDraw, tickHasOverrun, tickSkipsDraw);
    
  } // updateProgress()
  
} // class MainWindow
