/*
 *  MouseMonitor.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.event.*;

// monitor mouse behaviour within game component
public class MouseMonitor implements MouseListener, MouseMotionListener {

  // record of the state of the mouse pointer
  public static class State {
    public int x, y; // (in tile coordinates, or -1)
    public boolean b1, b2; // (buttons)
    public State(int xx, int yy, boolean bb1, boolean bb2) {
      x=xx; y=yy; b1=bb1; b2=bb2;
    }
  } // class MouseMonitor.State
  
  // current state of the pointer (position is in pixels)
  private int     mPointerX,
                  mPointerY;
  private boolean mButton1Pressed,
                  mButton2Pressed;
  
  // check for a click since the last reset
  private boolean mUnclicked,
                  mClicked;
  
  // constructor
  public MouseMonitor() {
    
    mPointerX = mPointerY = -1;
    mButton1Pressed = mButton2Pressed = false;
    
    mUnclicked = mClicked = false;
    
  } // constructor
  
  // monitor the mouse on this component
  public void monitor(Component target) {
    
    if ( target == null ) return;
    target.addMouseListener(this);
    target.addMouseMotionListener(this);
    
  } // monitor()
  
  // start waiting for a click
  public void reset() { 
    
    mUnclicked = (!mButton1Pressed && !mButton2Pressed); 
    mClicked = false;
    
  } // reset()
  
  // any click since last reset()
  public boolean clicked() { return mClicked; }
  
  // retrieve the current state of the pointer
  public synchronized State getState() {

    return new State(mPointerX, mPointerY, mButton1Pressed, mButton2Pressed);
    
  } // getState()
  
  // update current state of the pointer and the main button
  // (button: -1 => released, +1 => pressed, 0 => no change)
  private synchronized void updateState(int x, int y, 
                                        int button1, int button2) {
    
    if ( x >= 0 && x < Env.screenWidth() &&
         y >= 0 && y < Env.screenHeight() ) {
      mPointerX = x;
      mPointerY = y;
      if ( button1 != 0 ) mButton1Pressed = (button1 == +1);
      if ( button2 != 0 ) mButton2Pressed = (button2 == +1);
      
      if ( !mUnclicked && !mButton1Pressed && !mButton2Pressed ) {
        mUnclicked = true;
      } 
      if ( mUnclicked && (mButton1Pressed || mButton2Pressed) ) {
        mClicked = true;
      }
    } else {
      mPointerX = mPointerY = -1;
      mButton1Pressed = mButton2Pressed = false;
      reset();
    }
    
  } // updateState()
  
  // functions from MouseListener interface
  public void mouseClicked(MouseEvent e) {
    updateState(e.getX(), e.getY(), 0, 0);
  } // MouseListener.mouseClicked()
  public void mouseEntered(MouseEvent e) {
    updateState(e.getX(), e.getY(), -1, -1);
  } // MouseListener.mouseEntered()
  public void mouseExited(MouseEvent e) {
    updateState(-1, -1, -1, -1);
  } // MouseListener.mouseExited()
  public void mousePressed(MouseEvent e) {
    int b1 = ( (e.getButton()==MouseEvent.BUTTON1) ? +1 : 0 ),
        b2 = ( (e.getButton()==MouseEvent.BUTTON3) ? +1 : 0 );
    updateState(e.getX(), e.getY(), b1, b2);
  } // MouseListener.mousePressed()
  public void mouseReleased(MouseEvent e) {
    int b1 = ( (e.getButton()==MouseEvent.BUTTON1) ? -1 : 0 ),
        b2 = ( (e.getButton()==MouseEvent.BUTTON3) ? -1 : 0 );
    updateState(e.getX(), e.getY(), b1, b2);
  } // MouseListener.mouseReleased()
  
  // functions from MouseMotionListener interface
  public void mouseDragged(MouseEvent e) {
    updateState(e.getX(), e.getY(), 0, 0);
  } // MouseMotionListener.mouseDragged()
  public void mouseMoved(MouseEvent e) {
    updateState(e.getX(), e.getY(), 0, 0);
  } // MouseMotionListener.mouseMoved()
  
} // class MouseMonitor
