/*
 *  Resources.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.applet.Applet;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.imageio.ImageIO;
import javax.sound.sampled.*;

// manager for graphics, sounds, etc.
public class Resources {

  // path for files
  static protected final String kRootPackageName = "centipede/";
  
  // enumeration of resource labels
  static public final int kSprites         = 0,
                          kSymbolsCollages = 1,
                          kNumResources    = 2;
  // table of resource file names
  static protected final String[] kResourceFiles = { "contestSprites.gif", // 0
                                                     "symbols.clg" };      // 1
  
  // blank background colour
  static private final Color kBlankColour = new Color(0, 0, 0, 0);
  
  protected Vector[] mResources = null; // an array of Vectors of resources
  
  // constructor
  public Resources() {
    
    mResources = new Vector[kNumResources];
    
  } // constructor
  
  // open a stream to a resource file (return null if error)
  // (file is assumed to be within resources package)
  public InputStream openResourceStream(String fileName) {

    try {
      String resourceName = kRootPackageName + fileName;
      URL url = Resources.class.getClassLoader().getResource(resourceName);
      return url.openStream();
    } catch (Exception ex) {
      throw new RuntimeException("Resources.openResourceStream(): bad resource " 
                                 + fileName + ": " + ex.getMessage());
    }
  
  } // openResourceStream()

  // return a sprite sheet resource (exception if not known)
  public BufferedImage getSpriteSheet(int resourceNumber) {
    
    BufferedImage image = null;
    
    Vector resource = establishResource(resourceNumber);
    image = (BufferedImage)resource.get(0);
    
    return image;
    
  } // getSpriteSheet()
  
  // convert a sprite sheet to a compatible, translucent image
  protected BufferedImage prepareImage(BufferedImage imageIn) {
    
    int width  = imageIn.getWidth(),
        height = imageIn.getHeight();
    
    BufferedImage imageTemp = Env.createBitmaskImage(width, height);
    Graphics2D g2 = imageTemp.createGraphics();
    g2.setBackground(kBlankColour);
    g2.clearRect(0, 0, width, height);
    g2.drawImage(imageIn, 0, 0, null);
    g2.dispose();
    
    WritableRaster raster = imageTemp.getRaster();
    int pixels[] = ((DataBufferInt)(raster.getDataBuffer())).getData();
    int pixInd = 0;
    int maskVal = pixels[pixInd];
    for ( int j = 0 ; j < height ; j++ ) {
      for ( int i = 0 ; i < width ; i++ ) {
        if ( pixels[pixInd] == maskVal ) pixels[pixInd] = 0;
        pixInd++;
      }
    }
    
    BufferedImage imageOut = Env.createBitmaskImage(width, height);
    g2 = imageOut.createGraphics();
    g2.setBackground(kBlankColour);
    g2.clearRect(0, 0, width, height);
    g2.drawImage(imageTemp, 0, 0, null);
    g2.dispose();
        
    return imageOut;
    
  } // prepareImage()
  
  // extract a subset of a sprite sheet
  public BufferedImage getSprite(int resourceNumber, 
                                 int x, int y,
                                 int width, int height,
                                 boolean reverseX,
                                 boolean reverseY) {
    
    BufferedImage spriteSheet = getSpriteSheet(resourceNumber);
    
    int x1 = ( reverseX ? (x+width) : x ),
        x2 = ( reverseX ? x : (x+width) ),
        y1 = ( reverseY ? (y+height) : y ),
        y2 = ( reverseY ? y : (y+height) );
    
    BufferedImage image = Env.createBitmaskImage(width, height);
    Graphics2D g2 = image.createGraphics();
    g2.setBackground(kBlankColour);
    g2.clearRect(0, 0, width, height);
    g2.drawImage(spriteSheet, 
                 0, 0, width, height,
                 x1, y1, x2, y2, null);
    g2.dispose();
            
    return image;
    
  } // getSprite()
  
  /*
  // return a named Collage from a resource file (exception if not known)
  public Collage getCollage(int resourceNumber, String collageName) {
    
    // check the specified resource is available
    Vector resource = establishResource(resourceNumber);
    
    // search the resource for the named object
    for ( ListIterator it = resource.listIterator() ; it.hasNext() ; ) {
      Collage collage = (Collage)it.next();
      if ( collage.getName().equals(collageName) ) return collage;
    }
    
    // if not found, complain
    throw new RuntimeException("Resources.getCollage(): collage \"" 
                               + collageName + "\" not found in resource \"" 
                               + kResourceFiles[resourceNumber] + "\"");
    
  } // getCollage()
  */
  
  /*
  // return the contents of a sound file as a Clip object
  public Clip loadSoundClip(String fileName) throws IOException {
    
    Clip clip = null;

    try {
      
      String resourceName = kRootPackageName + fileName;
      URL url = Resources.class.getClassLoader().getResource(resourceName);
      
      if ( fileName.endsWith(".ogg") ) {

        InputStream inStream = url.openStream();
        clip = OggDecode.toClip(inStream);
        inStream.close();

      } else {

        AudioInputStream audioStream = AudioSystem.getAudioInputStream(url);
        clip = AudioSystem.getClip();
        clip.open(audioStream);
        audioStream.close();
        
      }
      
    } catch ( Exception ex ) {
      throw new IOException("Resources.loadSoundClip(): bad clip " 
                            + fileName + ": " + ex.getMessage());
    }
    
    return clip;
    
  } // loadSoundClip()
  */
  
  /*
  // return the contents of a sound file as a SoundEffect object
  public SoundEffect loadSoundEffect(String fileName) throws IOException {
    
    SoundEffect effect = null;
    
    try {
      
      String resourceName = kRootPackageName + fileName;
      URL url = Resources.class.getClassLoader().getResource(resourceName);
      
      if ( fileName.endsWith(".ogg") ) {

        InputStream inStream = url.openStream();
        effect = OggDecode.toSoundEffect(inStream);
        inStream.close();

      } else {

        AudioInputStream audioStream = AudioSystem.getAudioInputStream(url);
        Clip clip = AudioSystem.getClip();
        clip.open(audioStream);
        audioStream.close();
        effect = new SoundEffect(clip);
        
      }
      
    } catch ( Exception ex ) {
      throw new IOException("Resources.loadSoundEffect(): could not read " 
                            + fileName + ": " + ex.getMessage());
    }
    
    return effect;
    
  } // loadSoundEffect()
  */
  
  // check validity of specified resource, load if necessary, return
  protected Vector establishResource(int resourceNumber) {
    
    // check number is in enumeration
    if ( resourceNumber < 0 || resourceNumber >= kNumResources ) {
      throw new RuntimeException("Resources.establishResource(): "
                                 + "bad resource number (" 
                                 + resourceNumber + ")");
    }
    
    // load the resource vector if haven't already done so
    if ( mResources[resourceNumber] == null ) {

      String resourceName = kResourceFiles[resourceNumber];
      
      try {

        InputStream inS = openResourceStream(resourceName);
        BufferedInputStream inB = new BufferedInputStream(inS);
        
        if ( resourceName.endsWith(".clg") ) {
          // assume file contains a number of Collages
          //ObjectInputStream in = new ObjectInputStream(inB);
          //int numCollages = in.readInt();
          //mResources[resourceNumber] = new Vector(numCollages);
          //for ( int i = 0 ; i < numCollages ; i++ ) {
          //  mResources[resourceNumber].add(Collage.readCollage(in));
          //}
          //in.close();
        } else if ( resourceName.endsWith(".gif") ) {
          // assume file contains a sprite sheet
          mResources[resourceNumber] = new Vector(1);
          BufferedImage image = prepareImage(ImageIO.read(inB));
          mResources[resourceNumber].add(image);
          inB.close();
        } else {
          // assume file contains a serialized vector
          ObjectInputStream in = new ObjectInputStream(inB);
          mResources[resourceNumber] = (Vector)in.readObject();
          in.close();
        }
        
      } catch (Exception ex) {
        throw new RuntimeException("Resources.establishResource(): "
                                   + "error reading \"" + resourceName 
                                   + "\": " + ex.getClass().getName() 
                                   + " " + ex.getMessage());
      }
    }
    
    // return reference to resource
    return mResources[resourceNumber];
    
  } // establishResource()
  
} // class Resources
