/*
 *  Segment.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.LinkedList;

// a body segment
public class Segment {

  // images for body segments
  static private final int kNumBodyImages       = 6;
  static private final int kBodyImageWidth      = 27,
                           kBodyImageHeight     = 28;
  static private final int kBodyImageX          = 90,
                           kBodyImageY          = 33,
                           kBodyImageSeparation = 37;
  static private BufferedImage kBodyImages[] = null;
  static private BufferedImage kHighlightImages[] = null;
  
  // blank background colour
  static private final Color kBlankColour = new Color(0, 0, 0, 0);
  
  // collision region
  static private final int kCollisionX = 12,
                           kCollisionY = 13,
                           kCollisionR = 10;
  
  // rate of animation
  static private final int kAnimationTicks = 3;

  // height on screen
  static private final int kYPos = Background.groundLevel()-kBodyImageHeight;
  static private final int kMouseYMargin = 10;

  // parameters of boost
  static public final double kBoostSpeed = 4.0;
  static private final int kBoostDuration = 20;
  
  // parameters after boost
  static private final double kReturnSpeed = 2.0,
                              kReturnAccel = 0.1;
  
  // segment position
  private double mXPos;
  
  // current speed
  private double mXVel;
  
  // rest position
  private double mTargetPos;
  
  // remaining time at max speed
  private int mBoostTime;
  
  // counter for animation
  private int mTime;
  
  // load resources
  static public void initialize() {
  
    if ( kBodyImages != null ) return;
    
    kBodyImages = new BufferedImage[kNumBodyImages];
    for ( int k = 0 ; k < kNumBodyImages ; k++ ) {
      int x = kBodyImageX + k*kBodyImageSeparation;
      kBodyImages[k] = Env.resources().getSprite(Resources.kSprites, 
                                                 x, kBodyImageY, 
                                                 kBodyImageWidth, 
                                                 kBodyImageHeight,
                                                 false, false);
    }

    kHighlightImages = new BufferedImage[kNumBodyImages];
    for ( int k = 0 ; k < kNumBodyImages ; k++ ) {
      kHighlightImages[k] = recolouredImage(kBodyImages[k]);
    }
    
  } // initialize()
  
  // make a new image with different colours
  static private BufferedImage recolouredImage(BufferedImage imageIn) {
    
    int width  = imageIn.getWidth(),
        height = imageIn.getHeight();
    
    BufferedImage imageTemp = Env.createBitmaskImage(width, height);
    Graphics2D g2 = imageTemp.createGraphics();
    g2.setBackground(kBlankColour);
    g2.clearRect(0, 0, width, height);
    g2.drawImage(imageIn, 0, 0, null);
    g2.dispose();
    
    WritableRaster raster = imageTemp.getRaster();
    int pixels[] = ((DataBufferInt)(raster.getDataBuffer())).getData();
    int pixInd = 0;
    for ( int j = 0 ; j < height ; j++ ) {
      for ( int i = 0 ; i < width ; i++ ) {
        int val = pixels[pixInd];
        int a = (val >> 24) & 255,
            r = (val >> 16) & 255,
            g = (val >> 8) & 255,
            b = (val)      & 255;
        b = 255;
        pixels[pixInd] = ((a<<24) | (r<<16) | (g<<8) | (b) );
        pixInd++;
      }
    }
    
    BufferedImage imageOut = Env.createBitmaskImage(width, height);
    g2 = imageOut.createGraphics();
    g2.setBackground(kBlankColour);
    g2.clearRect(0, 0, width, height);
    g2.drawImage(imageTemp, 0, 0, null);
    g2.dispose();
        
    return imageOut;
    
    
  } // recolouredImage()
  
  // constructor for a segment in a new chain
  public Segment(int xPos) {
    
    initialize();
    
    mXPos = xPos;
    mTargetPos = mXPos;
    mXVel = 0.0;
    mBoostTime = 0;
    
    mTime = 0;
    
  } // constructor
  
  // constructor for a segment that's just landed
  public Segment(int xPos, double xVel) {
    
    initialize();
    
    mXPos = xPos;
    mTargetPos = Env.screenWidth()/2;
    mXVel = xVel;
    mBoostTime = 0;
    
    mTime = 0;
    
  } // constructor
  
  // accessors
  public double xPos() { return mXPos; }
  public double yPos() { return kYPos; }
  public double xVel() { return mXVel; }
  public int boostTime() { return mBoostTime; }
  static public int width() { return kBodyImageWidth; }
  
  // specifiy target position
  public void setRestPos(int x) { mTargetPos = x; }
  
  // specify both velocity and boost time
  public void setXVel(double xVel, int boostTime) {
    
    assert( xVel >= -kBoostSpeed && xVel <= +kBoostSpeed );
    assert( boostTime >= 0 && boostTime <= kBoostDuration );
    
    mXVel = xVel;
    mBoostTime = boostTime;
    
  } // setXVel()
  
  // impose time
  public void setTime(int t) { mTime = t; }
  
  // whether mouse if over segment
  public boolean mouseHit(int x, int y) {
    
    return ( x >= mXPos && 
             x <= mXPos + kBodyImageWidth &&
             y >= kYPos - kMouseYMargin &&
             y <= kYPos + kBodyImageHeight - kMouseYMargin );
    
  } // mouseHit()
  
  // distance of the mouse point from the segment (-1 for infinite)
  public int mouseDistance(int x, int y) {
    
    if ( y < kYPos - kMouseYMargin ||
         y >= kYPos + kBodyImageHeight + kMouseYMargin ) return -1;
    
    double dist = Math.abs((mXPos+kBodyImageWidth/2-1) - x);
    return (int)Math.round(dist);
    
  } // mouseDistance()
  
  // check for collision against a circle (returns true if so)
  public boolean collides(double x, double y, double r) {
  
    double x0 = mXPos + kCollisionX,
           y0 = kYPos + kCollisionY,
           r0 = kCollisionR;
    double dx = x0 - x,
           dy = y0 - y,
           d2 = dx*dx + dy*dy;
    return ( d2 < (r0+r)*(r0+r) );
    
  } // collides()
  
  // boost to the left (-1) or right (+1)
  public void boost(int direction) {
    
    assert( direction == -1 || direction == +1 );
    mXVel = kBoostSpeed*direction;
    mBoostTime = kBoostDuration;
    
  } // boostLeft()
  
  // update position, return true if touching minimum position
  public boolean advance(double minXPos) {

    boolean contact = false;
    
    mXPos += mXVel;
    if ( mXPos < minXPos ) {
      mXPos = minXPos;
      contact = true;
    }
    
    if ( mBoostTime > 0 ) {
      mBoostTime -= 1;
    } else {
      double targetVel = 0.0;
      if      ( mXPos > mTargetPos+kReturnSpeed ) targetVel = -kReturnSpeed;
      else if ( mXPos < mTargetPos-kReturnSpeed ) targetVel = +kReturnSpeed;
      if      ( mXVel > targetVel+0.5*kReturnAccel ) mXVel -= kReturnAccel;
      else if ( mXVel < targetVel-0.5*kReturnAccel ) mXVel += kReturnAccel;
      else                                           mXVel = targetVel;
    }
    
    mTime++;
    
    return contact;
    
  } // Sprite.advance()

  // display
  public void draw(Graphics2D g2, boolean highlighted) {
    
    int index = Env.fold(mTime, kAnimationTicks*kNumBodyImages)
                / kAnimationTicks;
    g2.drawImage((highlighted ? kHighlightImages[index] : kBodyImages[index]), 
                 (int)Math.round(mXPos), kYPos, null);
    
} // Sprite.draw()

  // test sprite images
  private void drawTest(Graphics2D g2) {
    
    AffineTransform oldTransform = g2.getTransform();
    AffineTransform at = AffineTransform.getScaleInstance(4,4);
    g2.setTransform(at);
    
    g2.setColor(Color.MAGENTA);
    
    for ( int k = 0 ; k < kNumBodyImages ; k++ ) {
      int x = 30*(k%3),
          y = 30*(k/3);
      g2.fillRect(x, y, kBodyImageWidth, kBodyImageHeight);
      g2.drawImage(kBodyImages[k], x, y, null);
    }
    
    g2.setTransform(oldTransform);
    
  } // drawTest()
  
} // class Segment
