/*
 *  TextImage.java
 *  Copyright Simon Hern 2008
 *  Contact: dishmoth@yahoo.co.uk, www.dishmoth.com
 */

package centipede;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.image.BufferedImage;
import java.util.*;

// a simple piece of text on the screen
// (also a static method for creating text images)
public class TextImage extends Sprite {

  // pixel gap around the text
  private static final int kMargin = 1;
  
  // non-colour
  private static final Color kBlankColour = new Color(0, 0, 0, 0);

  // font details
  private static final String kFontName = "SansSerif";
  private static final int    kFontType = Font.BOLD;
  
  // create a buffered image containing the specified text
  // (height is the final height of the image in pixels)
  static public BufferedImage createTextImage(String text, int height,
                                              Color colour, int fontSize) {
    
    FontRenderContext frc = new FontRenderContext(null, true, true);

    Font font = new Font(kFontName, kFontType, fontSize);
    Rectangle2D maxBounds = font.getMaxCharBounds(frc);
    int fontHeight = (int)Math.ceil(maxBounds.getHeight()),
        fontAscent = (int)Math.round(-maxBounds.getY());

    assert( height > 2*kMargin );
    double scale = (height - 2*kMargin) / (double)fontHeight;
    
    GlyphVector gvec = font.createGlyphVector(frc, text);
    Rectangle2D bounds = gvec.getLogicalBounds();
    int width = (int)Math.ceil(scale*bounds.getWidth() + 2*kMargin);
    
    BufferedImage image = Env.createTranslucentImage(width, height);
    
    Graphics2D g2 = (Graphics2D)image.getGraphics();
    g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                        RenderingHints.VALUE_ANTIALIAS_ON);
    g2.setRenderingHint(RenderingHints.KEY_FRACTIONALMETRICS,
                        RenderingHints.VALUE_FRACTIONALMETRICS_ON);

    g2.setBackground(kBlankColour);
    g2.clearRect(0, 0, width, height);

    g2.setTransform(AffineTransform.getScaleInstance(scale, scale));    
    g2.setColor(colour);
    g2.drawGlyphVector(gvec, (float)(kMargin/scale), 
                             (float)(kMargin/scale + fontAscent));

    g2.dispose();
    
    return image;
    
  } // createTextImage()
  
  // pixel position of text (y is centred, x is justified as below)
  private int mXPos,
              mYPos;
  
  // justifaction (-1 => left, 0 => centre, +1 => right)
  private int mJustification;

  // image of the text
  private BufferedImage mImage = null;
  
  // constructor
  public TextImage(String text, int x, int y, int height,
                   Color colour, int fontSize, int justification,
                   int screenLayer) {
    
    super(screenLayer);
    
    mXPos = x;
    mYPos = y;
    
    assert( justification >= -1 && justification <= +1 );
    mJustification = justification;
    
    mImage = createTextImage(text, height, colour, fontSize);
    
  } // constructor
  
  // advance currently does nothing
  public void advance(LinkedList<Sprite>     addTheseSprites, 
                      LinkedList<Sprite>     killTheseSprites,
                      LinkedList<StoryEvent> newStoryEvents) {
  } // Sprite.advance()
  
  // display the text
  public void draw(Graphics2D g2) {

    assert( mJustification >= -1 && mJustification <= +1 );
    int x = mXPos - (mJustification+1)*mImage.getWidth()/2,
        y = mYPos - mImage.getHeight()/2;
    g2.drawImage(mImage, x, y, null);
    
  } // Sprite.draw()
  
} // class TextImage
